static char *RCSid = "$Header: fairland.c,v 1.2 89/12/10 01:25:10 mr-frog Exp $";
/* do not change these 4 defines */
#define LANDMIN		1		/* plate altitude for normal land */
#define HILLMIN		34		/* plate altitude for hills */
#define PLATMIN		36		/* plate altitude for plateau */
#define HIGHMIN		98		/* plate altitude for mountains */

/* define ORE 1 to add resources, define ORE 0 if you want to use another
   program to add the resources */
#define ORE 1

/* raise FERT_MAX for more fertility */
#define FERT_MAX        (PLATMIN + (HIGHMIN - PLATMIN)/3)

/* raise OIL_MAX for more oil */
#define OIL_MAX         (HILLMIN - 1)

/* lower IRON_MIN for more iron */
#define IRON_MIN        ((2*HILLMIN)/3)

/* lower GOLD_MIN for more gold */
#define GOLD_MIN        (PLATMIN)

/* lower URAN_MIN for more uranium */
#define URAN_MIN        (PLATMIN + (HIGHMIN - PLATMIN)/3)

/*
fairland.c

version 1.6

This is a land generator for empire that creates a specified number of
continents, and spaces them evenly accross the map.  It tries to make
all of them the same size, and makes sure that they don't meet.  You
can get nearly any kind of map you want by choosing the right
arguments.

The empire file-io stuff I copied from the old land.c file written by James
W. Anderson (1987) and Dave Pare (1989).  I wrote the land generating part.

Here is the syntax:

fairland <nc> <sc> [<ni>] [<is>] [<sp>] [<pm>] [<di>] [<id>]

nc = number of continents
sc = continent size
ni = number of islands (optional, default = nc)
is = average size of islands (optional, default = sc/2)
sp = spike percentage: 0 = round, 100 = snake (default = 10)
pm = the percentage of land that is mountain (default = 0)
di = the minimum distance between continents (default = 2)
id = minimum distance from islands to continents (default = 1)

Note that all of the arguments but the first two are optional.

The algorithm starts out by placing the "capitols" on the torus in
such a way so as to maximize their distances from one another (this
uses the perterbation technique of calculus of variations).  Then from
these seeds, the continents are grown.  The kind of shape they grow
into is determined by the "spike" argument <sp>--the higher the spike,
the more spindly they will be.  If you lower the spike, the continents
will be more round.  The continents never touch one another, the
argument <di> specifies how many sectors of water should be between
the continents at all times.  The continents grow to the size you
specify with <sc> and they all get the same number of mountains
(specified by <pm>).  The other arguments should be self explanitory.
If #define ORE 1, then resources will be placed in the sectors as well.
You can alter the #define ORE_MAX, IRON_MIN, GOLD_MIN, FERT_MAX and 
URAN_MIN to affect what kind of resources the world gets.

Feedback is most welcome.

Ken Stevens
stevens@math.utoronto.ca

version 1.0 released 14 April 93

version 1.1 released 15 April 93
IMPROVEMENTS:
- The program now writes a script for automatically making all the countries
- Made the Airport markers an option (default turned off)
- Mountains are no longer put where capitols will be
- Improved algorithm for placing islands (less crowded)
- Changed the order of command line arguments -- most common are first
- Fixed the algorithm for finding coastal sectors (it will always find them)
- Re-order the continents so that the first players on will be closer
- Better map (show capitols and mountains)

version 1.2 released 17 April 93
IMPROVEMENTS:
- better elevation algorithm, so countries get even resources.

version 1.3 released 18 April 93
BUG FIX:
- removed the possibility of division by zero in elevate_land()

version 1.4 released 19 April 93
IMPROVEMENT:
- added #define DISTINCT_ISLANDS (default = 1) which ensures that islands do
  not crash into each other.

version 1.5 released 20 April 93
CHANGE:
- if #define ORE 1 (default = 1), then fairland will give the sectors
  resources.  The resources each sector gets is not random, but depends
  on its elevation.  Since fairland gives each continent exactly the same
  set of elevations, this means that all the continents will have exactly the
  same resources.  You can change the variables FERT_MAX, OIL_MAX, IRON_MIN,
  GOLD_MIN, and URAN_MIN to affect how many resources the continents get.

version 1.6 released 10 May 93
 IMPROVEMENTS:
- better iron, gold and uranium.
- fixed a bug that made continents 1 sector smaller than they should have been
- checks for capitols when placing resources so that the resources are more
  fairly distributed.
- changed default spike from 20% to 10%
- changed default % mountain from 10% to 0%
- fixed a bug which didn't display capitols when they wrapped around.
*/

#include <stdio.h>
#include <fcntl.h>
#include "var.h"
#include "misc.h"
#include "power.h"
#include "nat.h"
#include "sect.h"
#include "gamesdef.h"
#include "file.h"
#include "xy.h"

#define AIRPORT_MARKER 0 /* mark the continents with a * so you can tell them
			    from the islands 1 = mark, 0 = don't mark. */
#define DISTINCT_ISLANDS 1 /* don't let the islands crash into each other.
			      1 = don't merge, 0 = merge. */
#define XSIZE           ((WORLD_X) / 2) /* basically world x-y size */
#define YSIZE           (WORLD_Y)
#define STABLE_CYCLE 4       /* stability required for perterbed capitols */
#define INFINITY        999             /* a number which means "BIG" */

/* these defines prevent infinite loops:
*/

#define COAST_SEARCH_MAX 200  /* how many times do we look for a coast sector
				when growing continents and islands */
#define DRIFT_BEFORE_CHECK ((WORLD_X + WORLD_Y)/2)
#define DRIFT_MAX ((WORLD_X + WORLD_Y)*2)
#define MOUNTAIN_SEARCH_MAX 1000 /* how long do we try to place mountains */

/* handy macros:
*/

#define new_x(newx) (((newx) + WORLD_X) % WORLD_X)
#define new_y(newy) (((newy) + WORLD_Y) % WORLD_Y)
#define max(a,b) (a>b?a:b)
#define rnd(x) (random() % (x))

int sin;   /* number of sectors grown */
int ctot;  /* total number of continents and islands grown */
int *isin; /* array of how large each island is */

int nc, sc, di, sp, pm, ni, is, id; /* the 8 arguments to this program */
int *capx, *capy;                   /* location of the nc capitols */
int *mc, mcc = 0;                   /* array and counter used for stability
				       check when perturbing */
int spike;                          /* are we spiking? */
int mind;                           /* the final distance between capitols that
				       we achieved */
int dirx[] = {-2, -1, 1, 2, 1, -1}; /* gyujnb */
int diry[] = {0, -1, -1, 0, 1, 1};

int **own;                          /* owner of the sector.  -1 means water */
int **elev;                         /* elevation of the sectors */
int **sectx, **secty;               /* the sectors for each continent */
int **sectc;                        /* which sectors are on the coast? */
int *vector;                        /* used for measuring distances */
int *weight;                        /* used for placing mountains */
int *dsea, *dmoun;                  /* the dist to the ocean and mountain */
FILE *the_file;                     /* the file we write everything to */
struct	sctstr sects[YSIZE][XSIZE]; /* the data we write to that file */

/****************************************************************************
  MAIN
****************************************************************************/

main(argc, argv)
int argc;
char *argv[];
{
  parse_args(argc, argv);
  if (init() == -1)
    exit(-1);
  print_vars();
  puts("\n\n        #*# ...fairland rips open a rift in the datumplane... #*#\n\n");
  printf("placing capitals...\n");
  if (!drift())
    printf("fairland: unstable drift -- try increasing DRIFT_MAX\n");
  printf("growing continents...\n");
  grow_continents();
  printf("growing islands:");
  grow_islands();
  printf("\nelevating land...\n");
  create_elevations();
  printf("designating sectors...\n");
  if (ORE)
    printf("adding resources...\n");
  write_sects();
  printf("writing to sectors file...\n");
  if (write_file() == -1)
    exit (-1);
  output();
  write_newcap_script();
}

print_vars()
{
  puts("Creating a planet with:\n");
  printf("%d continents\n", nc);
  printf("continent size: %d\n", sc);
  printf("number of islands: %d\n", ni);
  printf("average size of islands: %d\n", is);
  printf("spike: %d%%\n", sp);
  printf("%d%% of land is mountain (each continent will have %d mountains)\n", pm, (pm*sc)/100);
  printf("minimum distance between continents: %d\n", di);
  printf("minimum distance from islands to continents: %d\n", id);
  printf("World dimensions: %dx%d\n", WORLD_X, WORLD_Y);
}

my_sqrt(n)
int n;     
{
  int i;

  for (i = 1; i*i < n*10000; ++i);
  return (i+50)/100;
}

/****************************************************************************
  PARSE COMMAND LINE ARGUMENTS
****************************************************************************/

parse_args(argc, argv)
int argc;
char *argv[];
{
  if (argc < 3 || argc > 9) {
    puts("fairland syntax:\n");
    puts("fairland <nc> <sc> [<ni>] [<is>] [<sp>] [<pm>] [<di>] [<id>]");
    puts("nc = number of continents");
    puts("sc = continent size");
    puts("ni = number of islands (default = nc)");
    puts("is = average size of islands (default = sc/2)");
    puts("sp = spike percentage: 0 = round, 100 = snake (default = 10)");
    puts("pm = the percentage of land that is mountain (default = 0)");
    puts("di = the minimum distance between continents (default = 2)");
    puts("id = minimum distance from islands to continents (default = 1)");
    exit (1);
  }
  nc = atoi(argv[1]);
  if (nc < 1) {
    puts("fairland: error -- number of continents must be >0");
    exit (1);
  }

  sc = atoi(argv[2]);
  if (sc < 1) {
    puts("fairland: error -- size of continents must be >0");
    exit (1);
  }

  if (argc > 3)
    ni = atoi(argv[3]);
  else
    ni = nc;

  if (argc > 4)
    is = atoi(argv[4]);
  else
    is = sc/2;
  if (is < 0)
    is = 0;

  if (argc > 5)
    sp = atoi(argv[5]);
  else
    sp = 10;
  if (sp < 0)
    sp = 0;
  if (sp > 100)
    sp = 100;

  if (argc > 6)
    pm = atoi(argv[6]);
  else
    pm = 0;
  if (pm < 0)
    pm = 0;

  if (argc > 7)
    di = atoi(argv[7]);
  else
    di = 2;

  if (di < 1) {
    puts("fairland: error -- distance between continents must be >0");
    exit (1);
  }
  if (di > WORLD_X/2 || di > WORLD_Y/2) {
    puts("fairland: error -- distance between continents too large");
    exit (1);
  }

  if (argc > 8)
    id = atoi(argv[8]);
  else
    id = 1;
  if (id < 1) {
    puts("fairland: error -- distance from islands to continents must be >0");
    exit (1);
  }
  if (id > WORLD_X || id > WORLD_Y) {
    puts("fairland: error -- distance from islands to continents too large");
    exit (1);
  }
  if (nc*sc + nc*my_sqrt(sc)*2*(di+1) > WORLD_X * WORLD_Y) {
    puts("fairland: error -- world not big enough to fit continents.");
    puts("arguments must satisfy:");
    puts("nc*sc*sc + nc*sqrt(sc)*2*(di+1) < WORLD_X * WORLD_Y");
    exit (1);
  }
}

/****************************************************************************
  VARIABLE INITIALIZATION
****************************************************************************/

init()
{
  int i, j, xx = 0, yy = 0;
  long now;

  the_file = (FILE *)open(empfile[EF_SECTOR].file, O_RDWR|O_CREAT|O_TRUNC, 0660);
  if (the_file < 0) {
    perror(empfile[EF_SECTOR].file);
    return -1;
  }
  time(&now);
  srandom(now+getpid());
  capx = (int *)calloc(nc, sizeof(int));
  capy = (int *)calloc(nc, sizeof(int));
  vector = (int *)calloc(WORLD_X+WORLD_Y, sizeof(int));
  mc = (int *)calloc(STABLE_CYCLE, sizeof(int));
  own = (int **)calloc(WORLD_X, sizeof(int*));
  elev = (int **)calloc(WORLD_X, sizeof(int*));
  for (i = 0; i < WORLD_X; ++i) {
    own[i] = (int *)calloc(WORLD_Y, sizeof(int));
    elev[i] = (int *)calloc(WORLD_Y, sizeof(int));
    for (j = 0; j < WORLD_Y; ++j) {
      own[i][j] = -1;
      elev[i][j] = -INFINITY;
    }
  }
  sectx = (int **)calloc(nc+ni, sizeof(int*));
  secty = (int **)calloc(nc+ni, sizeof(int*));
  sectc = (int **)calloc(nc+ni, sizeof(int*));
  isin = (int *)calloc(nc+ni, sizeof(int));
  weight = (int *)calloc(max(sc,is*2), sizeof(int));
  dsea = (int *)calloc(max(sc,is*2), sizeof(int));
  dmoun = (int *)calloc(max(sc,is*2), sizeof(int));
  for (i = 0; i < nc; ++i) {
    sectx[i] = (int *)calloc(sc, sizeof(int));
    secty[i] = (int *)calloc(sc, sizeof(int));
    sectc[i] = (int *)calloc(sc, sizeof(int));
  }
  for (i = nc; i < nc+ni; ++i) {
    sectx[i] = (int *)calloc(is*2, sizeof(int));
    secty[i] = (int *)calloc(is*2, sizeof(int));
    sectc[i] = (int *)calloc(is*2, sizeof(int));
  }

  for (i = 0; i < nc; ++i, xx += 2) {
    if (xx >= WORLD_X) {
      ++yy;
      xx = yy%2;
      if (yy == WORLD_Y) {
	puts("fairland error: world not big enough for all the continents.\n");
	exit(1);
      }
    }
    capx[i] = xx;
    capy[i] = yy;
  }
  for (i = 0; i < STABLE_CYCLE; ++i)
    mc[i] = i;
  return 0;
}
  
/****************************************************************************
  DRIFT THE CAPITOLS UNTIL THEY ARE AS FAR AWAY FROM EACH OTHER AS POSSIBLE
****************************************************************************/

/* How isolated is capitol j?
*/

iso(j, newx, newy)
int j, newx, newy;
{
  int i, md, d = WORLD_X + WORLD_Y;
  
  for (i = 0; i < nc; ++i) {
    if (i == j)
      continue;
    md = mapdist(capx[i], capy[i], newx, newy);
    if (md < d)
      d = md;
  }

  return d;
}

/* Drift all the capitols
*/

drift()
{
  int i, turns;
  
  for (turns = 0; turns < DRIFT_MAX; ++turns) {
    if (turns > DRIFT_BEFORE_CHECK && (mind = stable()))
      return 1;
    for (i = 0; i < nc; ++i)
      move(i);
  }
  return 0;
}

/* Check to see if we have stabilized--can we stop drifting the capitols?
*/

stable()
{
  int i, isod, d = 0, stab = 1;

  for (i = 0; i < nc; ++i) {
    isod = iso(i, capx[i], capy[i]);
    if (isod > d)
      d = isod;
  }
  for (i = 0; i < STABLE_CYCLE; ++i)
    if (d != mc[i])
      stab = 0;
  mc[mcc] = d;
  mcc = (mcc + 1) % STABLE_CYCLE;
  return stab?d:0;
}

/* This routine does the actual drifting
*/

move(j)
int j;
{
  int i, n, newx, newy;

  for (i = rnd(6), n = 0; n < 6; i = (i + 1)%6, ++n) {
    newx = new_x(capx[j] + dirx[i]);
    newy = new_y(capy[j] + diry[i]);
    if (iso(j, newx, newy) >= iso(j, capx[j], capy[j])) {
      capx[j] = newx;
      capy[j] = newy;
      return;
    }
  }
}

/****************************************************************************
  GROW THE CONTINENTS
****************************************************************************/

/* Look for a coastal sector of continent c
*/

find_coast(c)
int c;
{
  int i, j;
  
  for (i = 0; i < sin; ++i) {
    sectc[c][i] = 0;
    for (j = 0; j < 6; ++j)
      if (own[new_x(sectx[c][i] + dirx[j])][new_y(secty[c][i] + diry[j])] == -1)
	sectc[c][i] = 1;
  }
}

/* Used for measuring distances
*/

next_vector(n)
int n;
{
  int i;

  if (n == 1) {
    vector[0] += 1;
    vector[0] %= 6;
    return vector[0];
  }
  for (i = 1; i < n && vector[i] == vector[i-1]; ++i);
  vector[i-1] += 1;
  vector[i-1] %= 6;
  return i > 1 || vector[0] > 0;
}

/* Test to see if we're allowed to grow there: the arguments di and id
*/

try_to_grow(c, newx, newy, d)
int c, newx, newy, d;
{
  int i, j, px, py;

  for (i = 1; i <= d; ++i) {
    for (j = 0; j < i; ++j)
      vector[j] = 0;
    do {
      px = newx;
      py = newy;
      for (j = 0; j < i; ++j) {
	px = new_x(px + dirx[vector[j]]);
	py = new_y(py + diry[vector[j]]);
      }
      if (own[px][py] != -1 &&
	  own[px][py] != c &&
	  (DISTINCT_ISLANDS || own[px][py] < nc))
	return 0;
    } while (next_vector(i));
  }
  sectx[c][sin] = newx;
  secty[c][sin] = newy;
  own[newx][newy] = c;
  return 1;
}

/* Move along the coast in a clockwise direction.
*/

next_coast(c, x, y, xp, yp)
int c, x, y, *xp, *yp;
{
  int i, nx, ny, wat = 0;

  if (sin == 1) {
    *xp = x;
    *yp = y;
    return;
  }

  for (i = 0; i < 12; ++i) {
    nx = new_x(x+dirx[i%6]);
    ny = new_y(y+diry[i%6]);
    if (own[nx][ny] == -1)
      wat = 1;
    if (wat && own[nx][ny] == c) {
      *xp = nx;
      *yp = ny;
      return;
    }
  }
}

/* Choose a sector to grow from
*/

new_try(c)
{
  int i, starti;

  if (sin == 1) {
    if (sectc[c][0])
      return 0;
  }
  else {
    i = starti = (spike && sectc[c][sin-1])?sin-1:rnd(sin);
    do {
      if (sectc[c][i])
	return i;
      i = (i + 1)%sin;
    } while (i != starti);
    if (c < nc) {
      printf("fairland: BUG -- couldn't find coast for continent %c, sector %d.\nPlease mail stevens@math.utoronto.ca.\n", c + 'a', sin);
      exit (1);
    }
    else
      return -1;
  }
}

/* Grow continent c by 1 sector
*/

grow_one_sector(c)
int c;
{
  int done, coast_search, try, x, y, newx, newy, i, n, sx, sy;

  spike = rnd(100) < sp;
  if ((try = new_try(c)) == -1)
    return 0;
  x = sx = sectx[c][try];
  y = sy = secty[c][try];
  coast_search = 0;
  done = 0;
  do {
    if (spike) {
      for (i = rnd(6), n = 0; n < 12 && !done; i = (i + 1)%6, ++n) {
	newx = new_x(x+dirx[i]);
	newy = new_y(y+diry[i]);
	if (own[newx][newy] == -1 &&
	    (n > 5 ||
	     (own[new_x(x+dirx[(i+5)%6])][new_y(y+diry[(i+5)%6])] == -1 &&
	      own[new_x(x+dirx[(i+1)%6])][new_y(y+diry[(i+1)%6])] == -1)))
	  if (try_to_grow(c, newx, newy, c<nc?di:id))
	    done = 1;
      }
    }
    else
      for (i = rnd(6), n = 0; n < 6 && !done; i = (i + 1)%6, ++n) {
	newx = new_x(x+dirx[i]);
	newy = new_y(y+diry[i]);
	if (own[newx][newy] == -1)
	  if (try_to_grow(c, newx, newy, c<nc?di:id))
	    done = 1;
      }
    next_coast(c, x, y, &x, &y);
    ++coast_search;
  } while (!done && coast_search < COAST_SEARCH_MAX &&
	   (sin == 1 || x != sx || y != sy));
  if (!done && c < nc) {
    printf("fairland: error -- continent %c had no room to grow!\n", c+'a');    
    exit(1);
  }
  return done;
}

/* Grow all the continents
*/

grow_continents()
{
  int c;

  for (c = 0; c < nc; ++c) {
    sectx[c][0] = capx[c];
    secty[c][0] = capy[c];
    own[sectx[c][0]][secty[c][0]] = c;
    sectx[c][1] = new_x(capx[c] + 2);
    secty[c][1] = capy[c];
    own[sectx[c][1]][secty[c][1]] = c;
  }

  for (sin = 2; sin < sc; ++sin) {
    for (c = 0; c < nc; ++c) {
      find_coast(c);
      grow_one_sector(c);
    }
  }
  ctot = nc;
}

/****************************************************************************
  GROW THE ISLANDS
****************************************************************************/

/* Choose a place to start growing an island from
*/

place_island(c, xp, yp)
int *xp, *yp;
{
  int d, sx, sy;
  int ssy = rnd(WORLD_Y);
  int ssx = new_x(rnd(WORLD_X/2)*2 + ssy%2);

  if (ssx > WORLD_X - 2)
    ssx = new_x(ssx + 2);
  for (d = di + id; d >= id; --d) {
    sx = ssx;
    sy = ssy;
    *xp = new_x(sx + 2);
    for (*yp = sy; *xp != sx || *yp != sy; *xp += 2) {
      if (*xp >= WORLD_X) {
	*yp = new_y(*yp + 1);
	*xp = (*yp)%2;
	if (*xp == sx && *yp == sy)
	  break;
      }
      if (own[*xp][*yp] == -1 && try_to_grow(c, *xp, *yp, d))
	return 1;
    }
  }
  return 0;
}

/* Grow all the islands
*/

grow_islands()
{
  int c, x, y, isiz;

  for (c = nc; c < nc+ni; ++c) {
    sin = 0;
    if (!place_island(c, &x, &y))
      return;
    isiz = 1 + rnd(2*is - 1);
    do {
      ++sin;
      find_coast(c);
      } while (sin < isiz && grow_one_sector(c));
    printf(" %d(%d)", c - nc + 1, sin);
    isin[c] = sin;
    ctot = c;
  }
}

/****************************************************************************
  CREATE ELEVATIONS
****************************************************************************/

create_elevations()
{
  elevate_land();
  elevate_sea();
}

/* Generic function for finding the distance to the closest sea, land, or
   mountain
*/

distance_to_what(x,y,flag)
int x, y, flag;
{
  int j, d, px, py;

  for (d = 1; d < 5; ++d) {
    for (j = 0; j < d; ++j)
      vector[j] = 0;
    do {
      px = x;
      py = y;
      for (j = 0; j < d; ++j) {
	px = new_x(px + dirx[vector[j]]);
	py = new_y(py + diry[vector[j]]);
      }
      switch(flag) {
      case 0: /* distance to sea */
	if (own[px][py] == -1)
	  return d;
	break;
      case 1: /* distance to land */
	if (own[px][py] != -1)
	  return d;
	break;
      case 2: /* distance to mountain */
	if (elev[px][py] == INFINITY)
	  return d;
	break;
      }
    } while (next_vector(d));
  }
  return d;
}

#define ELEV elev[sectx[c][i]][secty[c][i]]
#define distance_to_sea() (sectc[c][i]?1:distance_to_what(sectx[c][i], secty[c][i], 0))
#define distance_to_mountain() distance_to_what(sectx[c][i], secty[c][i], 2)

/* Decide where the mountains go
*/

elevate_land()
{
  int i, mountain_search, k, c, total, ns, nm, highest, where, h, newk, r, dk;

  for (c = 0; c < ctot; ++c) {
    total = 0;
    ns = (c < nc)?sc:isin[c];
    nm = (pm*ns)/100;

/* Place the mountains */

    for (i = 0; i < ns; ++i) {
      dsea[i] = distance_to_sea();
      weight[i] = (total += (dsea[i]*dsea[i]));
    }
    
    for (k = nm, mountain_search = 0;
	 k && mountain_search < MOUNTAIN_SEARCH_MAX;
	 ++mountain_search) {
      r = rnd(total);
      for (i = 0; i < ns; ++i)
	if (r < weight[i] && ELEV == -INFINITY && (c >= nc || 
	    (!(capx[c] == sectx[c][i] && capy[c] == secty[c][i])) &&
	    (!(new_x(capx[c] + 2) == sectx[c][i] && capy[c] == secty[c][i])))) {
	  ELEV = INFINITY;
	  break;
	}
      --k;
    }

/* Elevate land that is not mountain and not capitol */

    for (i = 0; i < ns; ++i)
      dmoun[i] = distance_to_mountain();
    dk=(ns-nm-((c<nc)?3:1)>0)?(100*(HIGHMIN-LANDMIN))/(ns-nm-((c<nc)?3:1)):100*INFINITY;
    for (k = 100*(HIGHMIN - 1);; k -= dk) {
      highest = -INFINITY;
      where = -1;
      for (i = 0; i < ns; ++i) {
	if (ELEV != INFINITY && (c >= nc ||
            (!(capx[c] == sectx[c][i] && capy[c] == secty[c][i])) &&
            (!(new_x(capx[c] + 2) == sectx[c][i] && capy[c] == secty[c][i]))))
	  {
	    h = 3*(5 - dmoun[i]) + dsea[i];
	    if (h > highest) {
	      highest = h;
	      where = i;
	    }
	  }
      }
      if (where == -1)
	break;
      newk = k/100;
      if (newk >= HILLMIN && newk < PLATMIN)
	newk = PLATMIN;
      if (newk < LANDMIN)
	newk = LANDMIN;
      elev[sectx[c][where]][secty[c][where]] = newk;	
      dsea[where] = -INFINITY;
      dmoun[where] = INFINITY;
    }

/* Elevate the mountains and capitols */

    for (i = 0; i < ns; ++i) {
      if (ELEV == INFINITY) {
	if (dsea[i] == 1)
	  ELEV = HILLMIN + rnd(PLATMIN - HILLMIN);
	else
	  ELEV = HIGHMIN + rnd((256-HIGHMIN)/2) + rnd((256-HIGHMIN)/2);
      }
      else if (c < nc &&
            ((capx[c] == sectx[c][i] && capy[c] == secty[c][i])) ||
            ((new_x(capx[c] + 2) == sectx[c][i] && capy[c] == secty[c][i])))
	ELEV = PLATMIN;
    }
  }
}

#define distance_to_land() distance_to_what(x, y, 1)

elevate_sea()
{
  int x, y;

  for (y = 0; y < WORLD_Y; ++y) {
    for (x = y%2; x < WORLD_X; x +=2) {
      if (elev[x][y] == -INFINITY)
	elev[x][y] = - rnd((distance_to_land()*20 + 27)) - 1;
	}
  }
}

/****************************************************************************
  ADD THE RESOURCES
****************************************************************************/

set_fert(e)
int e;
{
  int fert = 0;
  if (e < LANDMIN)
    fert = LANDMIN - e + 40;
  else if (e < FERT_MAX)
    fert = (140*(FERT_MAX - e))/(FERT_MAX - LANDMIN);
  if (fert > 120)
    fert = 120;
  return fert;
}

set_oil(e)
int e;
{
  int oil = 0;
  if (e < LANDMIN)
    oil = (LANDMIN - e)*2 + rnd(2);
  else if (e <= OIL_MAX)
    oil = (120*(OIL_MAX - e + 1))/(OIL_MAX - LANDMIN + 1);
  if (oil > 100)
    oil = 100;
  return oil;
}

set_iron(e)
int e;
{
  int iron = 0;
  if (e >= IRON_MIN && e < HIGHMIN)
    iron = (120*(e - IRON_MIN + 1))/(HIGHMIN - IRON_MIN);
  if (iron > 100)
    iron = 100;
  return iron;
}

set_gold(e)
int e;
{
  int gold = 0;
  if (e >= GOLD_MIN && e < HIGHMIN)
    gold = (120*(e - GOLD_MIN + 1))/(HIGHMIN - GOLD_MIN);
  if (gold > 100)
    gold = 100;
  return gold;
}

set_uran(e)
int e;
{
  int uran = 0;
  if (e >= URAN_MIN && e < HIGHMIN)
    uran = (120*(e - URAN_MIN + 1))/(HIGHMIN - URAN_MIN);
  if (uran > 100)
    uran = 100;
  return uran;
}

add_resources(sct)
struct sctstr *sct;
{
  sct->sct_fertil = set_fert(sct->sct_elev);
  sct->sct_oil    = set_oil(sct->sct_elev);
  sct->sct_min    = set_iron(sct->sct_elev);
  sct->sct_gmin   = set_gold(sct->sct_elev);
  sct->sct_uran   = set_uran(sct->sct_elev);
}

/****************************************************************************
  DESIGNATE THE SECTORS
****************************************************************************/

write_sects()
{
  register struct sctstr *sct;
  int c, x, y, total;
  
  sct = &sects[0][0];
  for (y = 0; y < YSIZE; y++) {
    for (x = 0; x < XSIZE; x++, sct++) {
      sct->sct_x = x*2 + (y & 01);
      sct->sct_y = y;
      sct->sct_dist_x = x*2 + (y & 01);
      sct->sct_dist_y = y;
      total = elev[sct->sct_x][y];
      if (total < LANDMIN) {
	sct->sct_type = SCT_WATER;
      } else if (total < HILLMIN)
	sct->sct_type = SCT_RURAL;
      else if (total < PLATMIN)
	sct->sct_type = SCT_MOUNT;
      else if (total < HIGHMIN)
	sct->sct_type = SCT_RURAL;
      else
	sct->sct_type = SCT_MOUNT;
      sct->sct_elev = total;
      sct->sct_newtype = sct->sct_type;
      if (ORE)
	add_resources(sct);
    }
  }
  if (AIRPORT_MARKER)
    for (c = 0; c < nc; ++c) {
      sects[capy[c]][capx[c]/2 + capy[c]%2].sct_type = SCT_AIRPT;
      sects[capy[c]][capx[c]/2 + capy[c]%2].sct_newtype = SCT_AIRPT;
    }
}

/****************************************************************************
  WRITE ALL THIS STUFF TO THE FILE
****************************************************************************/

write_file()
{
  int n;

  if ((n = write(the_file, sects, sizeof(sects))) < 0) {
    perror(empfile[EF_SECTOR].file);
    return -1;
  }
  if (n != sizeof(sects)) {
    printf("%s:partial write\n", empfile[EF_SECTOR].file);
    return -1;
  }
  close(the_file);
  return 0;
}

/****************************************************************************
  PRINT A PICTURE OF THE MAP TO YOUR SCREEN
****************************************************************************/

output()
{
  int i, j;
  translate_continents();
  for (i = 0; i < WORLD_Y; ++i) {
    puts("");
    if (i%2)
      printf(" ");
    for (j = i%2; j < WORLD_X; j += 2) {
      if (own[j][i] == -1)
	printf(". ");
      else {
	printf("%c ", map_symbol(j,i));
      }
    }
  }
if (AIRPORT_MARKER) 
  printf("\n\nEach continent is marked by a \"*\" on the map (to distinguish them from\nthe islands).  You can redesignate these airfields to wilderness sectors\none at a time, each time you add a new country to the game.\n");
}

/* Reorder the continents from top left to bottom right */

translate_continents()
{
  int i, j, n = 0, k, gotit, c;
  int *trans, *trans_cont, *oldcapx, *oldcapy;

  trans = (int *)calloc(nc, sizeof(int));
  trans_cont = (int *)calloc(nc, sizeof(int));
  oldcapx = (int *)calloc(nc, sizeof(int));
  oldcapy = (int *)calloc(nc, sizeof(int));

  for (i = 0; i < WORLD_Y; ++i) {
    for (j = i%2; j < WORLD_X; j += 2) {
      if (own[j][i] >-1 && own[j][i] < nc) {
	gotit = 0;
	for (k = 0; k < n; ++k) {
	  if (trans[k] == own[j][i])
	    gotit = 1;
	}
	if (!gotit) {
	  if (n > nc) {
	    printf("fairland: BUG in translate continents!  mailstevens@math.utoronto.ca\n");
	    exit (2);
	  }
	  trans[n] = own[j][i];
	  trans_cont[own[j][i]] = n;
	  ++n;
	}
      }
    }
  }
  for (i = 0; i < WORLD_Y; ++i) {
    for (j = i%2; j < WORLD_X; j += 2) {
      if (own[j][i] >-1 && own[j][i] < nc) {
	own[j][i] = trans_cont[own[j][i]];
      }
    }
  }
  for (c = 0; c < nc; ++c) {
    oldcapx[c] = capx[c];
    oldcapy[c] = capy[c];
  }
  for (c = 0; c < nc; ++c) {
    capx[c] = oldcapx[trans[c]];
    capy[c] = oldcapy[trans[c]];
  }
}

map_symbol(x,y)
{
  int c, iscap = 0;

  for (c = 0; c < nc; ++c)
    if (x == capx[c] && y == capy[c] || x == new_x(capx[c]+2) && y == capy[c])
      iscap = 1;
  if (elev[x][y] >= HILLMIN && elev[x][y] < PLATMIN || elev[x][y] >= HIGHMIN)
    return '^';
  return own[x][y]>=nc?'X':iscap?'#':own[x][y]+'a';
}

/***************************************************************************
  WRITE A SCRIPT FOR PLACING CAPITOLS
****************************************************************************/

write_newcap_script()
{
  int c;
  FILE *script = fopen("newcap_script", "w");

  if (!script) {
    printf("fairland: error, unable to write to newcap_script.\n");
    return -1;
  }

  for (c = 0; c < nc; ++c) {
    fprintf(script, "add %d %d %d n i\n", c+1, c+1, c+1);
    if (AIRPORT_MARKER)
      fprintf(script, "des %d,%d -\n", capx[c], capy[c]);
    fprintf(script, "newcap %d %d,%d\n", c+1, capx[c], capy[c]);
  }
  printf("\n\nA script for adding all the countries can be found in \"newcap_script\".\n");
  if (ORE)
    printf("***IMPORTANT: Resources have already been added***\n              You do NOT need to run the ore program.\n");
  return 0;
}
